"""
merakitools - formatting_helpers.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
import re
from typing import List, Optional
from rich.table import Table, Column
from rich import box

# Mapping of styles to severity - TODO: move to an import
severity_styles = {"critical": "bold red", "warning": "yellow"}


def camel_case_split(input_string) -> str:
    """
    Input: applicationCategory
    Output: Application Category
    """
    return re.sub(r"(\w)([A-Z])", r"\1 \2", input_string).capitalize()


def table_with_columns(
    columns: List, title: Optional[str] = None, first_column_name: Optional[str] = None
) -> Table:
    """
    Generate a table with specified columns
    """
    table = Table(*columns, title=title, expand=False, box=box.ROUNDED)
    if first_column_name:
        first_column = Column(first_column_name, style="bold blue")
        table.columns.insert(0, first_column)

    return table


def table_mx_onetoone_nat(rules: List, title: str = "NAT Entries") -> Table:
    """
    Generate a table to display MX one to one NAT entries
    """
    table = table_with_columns(
        ["External IP", "Internal IP", "Uplink", "Protocol", "Ports", "Allowed IPs"],
        title=title,
        first_column_name="Name",
    )
    for rule in rules:
        table.add_row(
            rule["name"],
            rule["publicIp"],
            rule["lanIp"],
            rule["uplink"],
            "",
            "",
            "",
            style="bold",
        )
        for idx, entry in enumerate(rule["allowedInbound"]):
            is_last_row = idx == len(rule["allowedInbound"]) - 1
            table.add_row(
                "",
                "",
                "",
                "",
                entry["protocol"],
                ",".join(entry["destinationPorts"]),
                ",".join(entry["allowedIps"]),
                end_section=is_last_row,
            )

    return table


def table_network_health(health, title=None, include_network_name=False):
    """
    Create table of network health alerts
    """
    columns = ["Alert", "Category", "Severity", "Details"]
    if include_network_name:
        columns.insert(0, "Network")
        empty_columns = ("", "", "", "")
    else:
        empty_columns = ("", "", "")
    table = table_with_columns(columns, title=title)

    for alert in health:
        row_items = (
            alert["type"],
            alert["category"],
            f"[{severity_styles.get(alert['severity'], '')}]{alert['severity'].capitalize()}",
            "",
        )
        if include_network_name:
            table.add_row(alert["network_name"], *row_items)
        else:
            table.add_row(*row_items)

        # Create additional rows for devices / applications if needed
        for app in alert["scope"]["applications"]:
            table.add_row(*empty_columns, f"-- Application details --")

        for device in alert["scope"]["devices"]:
            detail = f"{device['productType'].capitalize()}: {device['name']}"
            if device.get("lldp"):
                detail += f" Port #{device['lldp']['portId']}"
            table.add_row(*empty_columns, detail)

    return table



"""
merakitools - ms.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
from typing import List, Optional
import time
from meraki.exceptions import APIError
import typer
from rich import inspect
from rich.progress import Progress, track
from merakitools.console import console, status_spinner
from merakitools.dashboardapi import dashboard
from merakitools.meraki_helpers import (
    find_network_by_name,
    find_org_id_by_device_serial,
)
from merakitools.formatting_helpers import table_with_columns
from merakitools.types import (
    DeviceModel,
    MSInterfaceMode,
    TrafficDirection,
    MSSTPGuardType,
)

app = typer.Typer()


@app.command()
def update_switchport(
    serial: str,
    port: Optional[List[int]] = typer.Option(None, min=1, max=52),
    port_range: Optional[str] = None,
    name: Optional[str] = None,
    enabled: Optional[bool] = None,
    poe_enabled: Optional[bool] = None,
    type: Optional[MSInterfaceMode] = None,
    vlan: Optional[int] = typer.Option(None, min=1, max=4094),
    voice_vlan: Optional[int] = typer.Option(None, min=1, max=4094),
    rtsp_enabled: Optional[bool] = None,
    stp_guard: Optional[MSSTPGuardType] = None,
    add_tag: Optional[List[str]] = None,
    remove_tag: Optional[List[str]] = None,
):
    """
    Update switchport(s)
    """
    # Validate device by trying to find OrgID
    org_id = find_org_id_by_device_serial(serial=serial)

    # Create range of ports
    ports = []
    if port_range:
        range_begin, range_end = port_range.split(":")
        ports = list(range(int(range_begin), int(range_end) + 1))

    # Add specified ports to the range
    for port in port:
        if port not in ports:
            ports.append(port)

    if len(ports) < 1:
        console.print("[red]No ports specified.")
        raise typer.Abort()

    # Iterate through each port in the range
    actions = []
    for port in track(ports, description="Processing switchports", console=console):
        try:
            port = dashboard.switch.getDeviceSwitchPort(serial=serial, portId=port)
        except APIError as err:
            console.print(err.message)
            raise typer.Abort()

        # Compile a dict of settings to be updated
        update = {}
        items = {
            "name": name,
            "enabled": enabled,
            "poeEnabled": poe_enabled,
            "type": type.value if type is not None else type,
            "vlan": vlan,
            "voiceVlan": voice_vlan,
            "rtspEnabled": rtsp_enabled,
            "stpGuard": stp_guard.value if stp_guard is not None else stp_guard,
        }
        for key, value in items.items():
            if value is not None:
                if value is not port[key]:
                    update[key] = value

        # Update tags
        if add_tag or remove_tag:
            update["tags"] = port["tags"]
            for tag in add_tag:
                if tag not in update["tags"]:
                    update["tags"].append(tag)
            for tag in remove_tag:
                if tag in update["tags"]:
                    update["tags"].remove(tag)

        # Create an action to be included in action batch
        if update:
            update_action = dashboard.batch.switch.updateDeviceSwitchPort(
                serial=serial, portId=port["portId"], **update
            )
            actions.append(update_action)

    # Use an action batch to execute in one run
    if actions:
        with status_spinner("Waiting for API batch to complete"):
            action_batch = dashboard.organizations.createOrganizationActionBatch(
                organizationId=org_id, actions=actions, confirmed=True
            )
            while True:
                if (
                    action_batch["status"]["completed"]
                    or action_batch["status"]["failed"]
                ):
                    break
                time.sleep(5)
                action_batch = dashboard.organizations.getOrganizationActionBatch(
                    organizationId=org_id, actionBatchId=action_batch["id"]
                )
        if action_batch["status"]["completed"]:
            console.print(f"[green]Updated {len(action_batch['actions'])} switchports")
            raise typer.Exit()
        console.print(f"[red]Failed to update switchports")
    else:
        console.print(f"[red]No changes to apply.")


@app.command()
def list_stacks(organization_name: str, network_name: str):
    """
    List switch stacks
    """
    # Get a list of switch stacks for specified network
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting switch stacks"):
        stacks = dashboard.switch.getNetworkSwitchStacks(networkId=net["id"])

    # Exit if no stacks are found
    if len(stacks) == 0:
        console.print("No switch stacks found.")
        raise typer.Exit()

    # Create a table of stacks
    table = table_with_columns(
        ["Serials"],
        title=f"Switch stacks in {net['name']}",
        first_column_name="Name",
    )
    for stack in stacks:
        table.add_row(stack["name"], f"{', '.join(stack['serials'])}")
    console.print(table)


@app.command()
def list_routing_interfaces(
    organization_name: str, network_name: str, serial: str, include_dhcp: bool = False
):
    """
    List L3 routed interfaces on an MS switch or stack
    """
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting routing interfaces"):
        try:
            routing_interfaces = dashboard.switch.getDeviceSwitchRoutingInterfaces(
                serial=serial
            )
            stack = False
        except APIError as err:
            if "switches in switch stack" in err.message["errors"][0].lower():
                console.print(f"This switch is a member of a stack.")
                stacks = dashboard.switch.getNetworkSwitchStacks(networkId=net["id"])
                stack = next(stack for stack in stacks if serial in stack["serials"])
                routing_interfaces = (
                    dashboard.switch.getNetworkSwitchStackRoutingInterfaces(
                        networkId=net["id"], switchStackId=stack["id"]
                    )
                )
            else:
                console.print(err.message)
                raise typer.Abort()

    if not routing_interfaces:
        console.print("No routing interfaces found.")
        raise typer.Exit()

    # Create and print a table
    cols = ["Subnet", "Interface IP", "VLAN ID"]
    if include_dhcp:
        cols.append("DHCP")
    table = table_with_columns(
        cols, title="Routing interfaces", first_column_name="Name"
    )
    for intf in routing_interfaces:
        rows = [intf["name"], intf["subnet"], intf["interfaceIp"], str(intf["vlanId"])]

        if include_dhcp:
            # Get DHCP info
            with status_spinner("Getting DHCP settings"):
                if stack:
                    dhcp = dashboard.switch.getNetworkSwitchStackRoutingInterfaceDhcp(
                        networkId=net["id"],
                        switchStackId=stack["id"],
                        interfaceId=intf["interfaceId"],
                    )
                else:
                    dhcp = dashboard.switch.getDeviceSwitchRoutingInterfaceDhcp(
                        serial=serial, interfaceId=intf["interfaceId"]
                    )

            # Format DHCP cell, and add to table rows
            if dhcp["dhcpMode"] == "dhcpDisabled":
                dhcp_formatted = "Disabled"
            elif dhcp["dhcpMode"] == "dhcpServer":
                if dhcp["dnsNameserversOption"] == "custom":
                    dhcp_formatted = (
                        "[bold]Server[/bold] DNS:"
                        f" {', '.join(dhcp['dnsCustomNameservers'])}"
                    )
                else:
                    dhcp_formatted = (
                        "[bold]Server[/bold]"
                        f" {dhcp['dnsNameserversOption'].capitalize()}"
                    )
            elif dhcp["dhcpMode"] == "dhcpRelay":
                dhcp_formatted = (
                    "[bold]Relay[/bold] Servers:"
                    f" {', '.join(dhcp['dhcpRelayServerIps'])}"
                )
            else:
                dhcp_formatted = dhcp["dhcpMode"].capitalize()
                inspect(dhcp)
            rows.append(dhcp_formatted)

        table.add_row(*rows)

    console.print(table)


@app.command()
def diag_switchport_traffic(
    organization_name: str,
    network_name: str,
    top: int = 100,
    sort_by: TrafficDirection = TrafficDirection.total,
    interface_mode: MSInterfaceMode = None,
    ignore_device_tag: List[str] = None,
    ignore_switchport_tag: List[str] = None,
):
    """
    Diagnose all switchports on a network of MS switches to find
    switchports that are top talkers at a given instant

    ** This command gathers a significant amount of data and may
    take some time to complete on larger networks **
    """
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting network devices"):
        devices = dashboard.networks.getNetworkDevices(net["id"])

    console.print(
        "Analyzing each switchport on the network, this may take a few minutes"
    )
    switchports = []
    with Progress(console=console) as progress:
        task_devices = progress.add_task(
            f"[blue]Processing {len(devices)} devices..", total=len(devices)
        )
        # Iterate through each switchport and add to the master switchport list
        for dev in devices:
            progress.update(
                task_devices,
                advance=1,
                description=f"[blue] Processing device '{dev['name']}'",
            )

            # Ignore non-MS devices
            if DeviceModel.MS not in dev["model"]:
                continue

            # Ignore devices with specific tags
            if ignore_device_tag and any(
                tag in dev["tags"] for tag in ignore_device_tag
            ):
                continue

            # Get switchport configuration and status
            switchport_configs = dashboard.switch.getDeviceSwitchPorts(
                serial=dev["serial"]
            )
            switchport_statuses = dashboard.switch.getDeviceSwitchPortsStatuses(
                serial=dev["serial"]
            )

            # Iterate through each switchport
            for swp_stat in switchport_statuses:
                # Ignore non-connected switchports
                if not swp_stat["status"] == "Connected":
                    continue

                # Create a combined switchport record (status+config)
                swp_cfg = next(
                    swp_cfg
                    for swp_cfg in switchport_configs
                    if swp_cfg["portId"] == swp_stat["portId"]
                )
                swp = swp_stat | swp_cfg
                swp["switch_name"] = dev["name"]

                # Ignore switchports with specific tags
                if ignore_switchport_tag and any(
                    tag in swp["tags"] for tag in ignore_switchport_tag
                ):
                    continue

                # Add combined switchport record to list of switchports
                switchports.append(swp)

        progress.update(
            task_devices, description=f"[green]Processed {len(devices)} devices."
        )

        # Sort all switchports based on the specified sort_by value (highest to lowest)
        switchports = sorted(
            switchports, key=lambda k: k["trafficInKbps"][sort_by.value], reverse=True
        )
        table = table_with_columns(
            [
                "Port",
                "Status",
                "Traffic (Total)",
                "Traffic (Sent)",
                "Traffic (Received)",
                "# of Clients",
                "CDP/LLDP",
                "Errors/Warnings",
            ],
            title=f"Top {top} switchports",
            first_column_name="Device",
        )

        for swp in switchports[:top]:
            # get Mbps values for traffic
            mb_size = 1024
            rounding_precision = 2
            total_mbps = round(
                swp["trafficInKbps"]["total"] / mb_size, rounding_precision
            )
            recv_mbps = round(
                swp["trafficInKbps"]["recv"] / mb_size, rounding_precision
            )
            sent_mbps = round(
                swp["trafficInKbps"]["sent"] / mb_size, rounding_precision
            )

            # get discovery data if available
            if "cdp" in swp.keys():
                discovery = swp["cdp"].get("systemName", None)
            elif "lldp" in swp.keys():
                discovery = swp["lldp"].get("systemName", None)
            else:
                discovery = None

            # collate error/warning data
            errorwarn = swp["errors"] + swp["warnings"]

            table.add_row(
                f"{swp['switch_name']}",
                f"{swp['portId']} ({swp['name']})",
                f"{swp['status']} ({swp['speed']})",
                f"{str(total_mbps)} Mbps",
                f"{str(sent_mbps)} Mbps",
                f"{str(recv_mbps)} Mbps",
                str(swp["clientCount"]),
                discovery,
                (", ").join(errorwarn),
            )

        console.print(table)



"""
merakitools - mx.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
from typing import List, Optional
import ipaddress
import typer
from rich.prompt import Confirm
from merakitools.console import console, status_spinner
from merakitools.dashboardapi import dashboard, APIError
from merakitools.meraki_helpers import (
    find_network_by_name,
)
from merakitools.formatting_helpers import table_with_columns, table_mx_onetoone_nat
from merakitools.types import MXInternetUplinks


app = typer.Typer()


@app.command()
def list_vlans(
    organization_name: str,
    network_name: str,
    include_dhcp: Optional[bool] = typer.Option(
        False, help="Include DHCP information for each subnet"
    ),
):
    # Find network and VLANs
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting VLANs"):
        if dashboard.appliance.getNetworkApplianceVlansSettings(networkId=net["id"])[
            "vlansEnabled"
        ]:
            vlans = dashboard.appliance.getNetworkApplianceVlans(networkId=net["id"])
        else:
            vlans = dashboard.appliance.getNetworkApplianceSingleLan(
                networkId=net["id"]
            )

    # Create a table, including extra columns if --include-dhcp is specified
    columns = ["ID", "Subnet", "Appliance IP"]
    if include_dhcp:
        columns = columns + ["DHCP", "Details"]
    table = table_with_columns(columns, first_column_name="Name")
    for vlan in vlans:
        cols = [
            vlan["name"],
            str(vlan["id"]),
            vlan["subnet"],
            vlan["applianceIp"],
        ]
        if include_dhcp:
            dhcp_cols = [vlan["dhcpHandling"]]
            if "Relay" in vlan["dhcpHandling"]:
                dhcp_cols.append(f"Relays: {', '.join(vlan['dhcpRelayServerIps'])}")
            elif "DHCP server" in vlan["dhcpHandling"]:
                dhcp_cols.append(f"Lease time: {vlan['dhcpLeaseTime']}")
            else:
                dhcp_cols.append("")

        table.add_row(*cols + dhcp_cols)
    console.print(table)


@app.command()
def list_routes(
    organization_name: str,
    network_name: str,
    name: Optional[str] = typer.Option(
        None, help="Only show routes with a specific name"
    ),
    subnet: Optional[str] = typer.Option(
        None, help="Only show routes with a specific subnet"
    ),
    gateway: Optional[str] = typer.Option(
        None, help="Only show routes with a specific gateway"
    ),
):
    """
    List MX device routes
    """
    # Find network and get static routes
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting routes"):
        routes = dashboard.appliance.getNetworkApplianceStaticRoutes(
            networkId=net["id"]
        )

    # Print number of routes, exit if no routes
    console.print(f"Found {len(routes)} total routes.")
    if not routes:
        raise typer.Abort()

    # Generate table, filtering if needed
    table = table_with_columns(
        ["Subnet", "Gateway", "Enabled"], first_column_name="Name"
    )
    for route in routes:
        # Filter by name
        if name is not None:
            if name not in route["name"]:
                continue

        # Filter by subnet
        if subnet is not None:
            if subnet not in route["subnet"]:
                continue

        # Filter by gateway
        if gateway is not None:
            if gateway not in route["gatewayIp"]:
                continue

        # Add to table
        table.add_row(
            route["name"],
            route["subnet"],
            route["gatewayIp"],
            "[green]Enabled" if route["enabled"] else "[red]Disabled",
        )
    console.print(table)


@app.command()
def add_staticroute(
    organization_name: str,
    network_name: str,
    route: Optional[List[str]] = None,
    default_next_hop: Optional[str] = None,
):
    """
    Add one or more static routes to an MX device
    """
    if not route:
        console.print("No routes provided.")
        raise typer.Abort()

    # Find the network
    net = find_network_by_name(organization_name, network_name)

    # Iterate through each provided route
    for route in route:
        args = len(route.split("!"))
        # Each route item should be formatted as 'routeCIDR' OR 'routeName!routeCIDR' or 'routeName!routeCIDR!routeNextHop'
        if not 0 < args < 4:
            console.print("Incorect --route formatting")
            raise typer.Abort()

        if args == 3:
            # 'routeName!routeCIDR!routeNextHop' format
            route_name, route_subnet, route_nexthop = route.split("!")
        elif default_next_hop:
            route_nexthop = default_next_hop
            if args == 1:
                # 'routeCIDR' format
                route_name = route.replace(".", "-").replace("/", "-")
                route_subnet = route
            elif args == 2:
                # 'routeName!routeCIDR' format
                route_name, route_subnet = route.split("!")
        else:
            console.print(
                "A next hop must be provided for each route, or --default-next-hop"
                " specified"
            )
            raise typer.Abort

        # Check route_subnet and route_nexthop validity
        try:
            route_subnet = ipaddress.ip_network(route_subnet)
        except ValueError as err:
            console.print(f"Invalid next hop IP {route_subnet}. {err}")
            raise typer.Abort()

        try:
            route_nexthop = ipaddress.ip_address(route_nexthop)
        except ValueError as err:
            console.print(f"Invalid next hop IP {route_nexthop}. {err}")
            raise typer.Abort()

        # Add route via API
        with status_spinner("Adding static route"):
            try:
                dashboard.appliance.createNetworkApplianceStaticRoute(
                    networkId=net["id"],
                    name=route_name,
                    subnet=format(route_subnet),
                    gatewayIp=format(route_nexthop),
                )
                console.print(
                    f"Created new route [bold]{route_name}[/bold] with subnet"
                    f" {route_subnet}"
                )
            except APIError as err:
                console.print(f"Failed to create route [bold]{route_name}[/bold]:")
                for e in err.message["errors"]:
                    console.print(f" - {e}")


@app.command()
def create_staticnat(
    organization_name: str,
    network_name: str,
    nat: Optional[List[str]] = None,
    port: Optional[List[str]] = None,
    uplink: MXInternetUplinks = MXInternetUplinks.one,
    confirm: bool = True,
):
    """
    Add a 1:1 NAT entry to an MX device
    """
    if not nat:
        console.print("No NATs provided.")
        raise typer.Abort()

    # Get a copy of existing NAT rules
    net = find_network_by_name(organization_name, network_name)
    rules = dashboard.appliance.getNetworkApplianceFirewallOneToOneNatRules(net["id"])[
        "rules"
    ]

    # Display a table of existing rules
    table = table_mx_onetoone_nat(rules, title="Existing 1:1 NAT Rules")
    console.print(table)

    # Iterate through each provided port
    allowed_ports = []
    for p in port:
        args = len(p.split("!"))
        # Each port item should  be formatted as 'protocol!portNum' or 'protocol!portNum!allowedIPs'
        if not 1 < args < 4:
            console.print("Incorect --port formatting")
            raise typer.Abort()

        if args == 3:
            # 'protocol!portNum!allowedIPs' format
            proto, num, allowed_ips = p.split("!")
        elif args == 2:
            # 'protocol!portNum' format
            proto, num = p.split("!")
            allowed_ips = "any"

        # Validate protocol input
        valid_protocols = ["tcp", "udp", "icmp", "any"]
        if not proto in valid_protocols:
            console.print(
                "Incorect --port protocol. Protocol must be"
                f" \[{', '.join(valid_protocols)}]"
            )
            raise typer.Abort()

        # Validate allowed IP input
        for ip in allowed_ips.split(","):
            if not ip == "any":
                try:
                    ip = ipaddress.ip_network(ip)
                except ValueError as err:
                    console.print(f"Invalid allowed IP {ip}. {err}")
                    raise typer.Abort()

        # Add to list of allowed_ports for all NAT entries
        allowed_ports.append(
            {
                "protocol": proto,
                "destinationPorts": num.split(","),
                "allowedIps": allowed_ips.split(","),
            }
        )

    # Iterate through each provided NAT
    new_rules = []
    for entry in nat:
        args = len(entry.split("!"))
        # Each NAT item should be formatted as 'publicIP!privateIP' or 'natName!publicIP!privateIP'
        if not 1 < args < 4:
            console.print("Incorect --nat formatting")
            raise typer.Abort()

        if args == 3:
            # 'natName!publicIP!privateIP' format
            name, public_ip, private_ip = entry.split("!")
        elif args == 2:
            # 'publicIP!privateIP' format
            public_ip, private_ip = entry.split("!")
            name = public_ip.replace(".", "-")

        # Validate public IP input
        try:
            public_ip = ipaddress.ip_address(public_ip)
        except ValueError as err:
            console.print(f"Invalid public IP {public_ip}. {err}")
            raise typer.Abort()

        # Validate private IP input
        try:
            private_ip = ipaddress.ip_address(private_ip)
        except ValueError as err:
            console.print(f"Invalid private IP {private_ip}. {err}")
            raise typer.Abort()

        # Add to list of new_rules
        new_rules.append(
            {
                "name": name,
                "lanIp": format(private_ip),
                "publicIp": format(public_ip),
                "uplink": uplink.value,
                "allowedInbound": allowed_ports,
            }
        )

    # Display a table of new rules
    table = table_mx_onetoone_nat(new_rules, title="New 1:1 NAT Rules")
    console.print(table)

    # Confirm before adding changes
    if confirm:
        request_confirm = Confirm.ask(
            "Do you want to continue adding new rules?", console=console
        )
        if not request_confirm:
            console.print("Rules not added")
            raise typer.Abort()

    rules += new_rules
    dashboard.appliance.updateNetworkApplianceFirewallOneToOneNatRules(net["id"], rules)



"""
merakitools - devices.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
from typing import List, Optional
from time import sleep
import typer
from rich import box
from rich.prompt import Confirm
from rich.table import Table
from merakitools.console import console, status_spinner
from merakitools.dashboardapi import dashboard
from merakitools.meraki_helpers import find_network_by_name, find_org_by_name
from merakitools.formatting_helpers import table_with_columns
from merakitools.types import DeviceModel, DeviceSortOptions

app = typer.Typer()


@app.command()
def list(
    organization_name: str,
    network_name: str,
    type: Optional[DeviceModel] = None,
    sort_by: Optional[DeviceSortOptions] = DeviceSortOptions.model,
    sort_reverse: bool = False,
):
    """
    List Meraki devices
    """
    # Get a list of all devices within the organization/network
    net = find_network_by_name(organization_name, network_name)
    devices = dashboard.networks.getNetworkDevices(net["id"])

    # Filter and sort
    if type:
        devices = [dev for dev in devices if type in dev["model"]]
    devices = sorted(devices, key=lambda k: k[sort_by.value], reverse=sort_reverse)

    # Display a table
    table = table_with_columns(
        ["Serial", "Network", "model", "tags", "Firmware"],
        title=f"Devices in {net['name']}",
        first_column_name="Name",
    )
    for device in devices:
        table.add_row(
            device["name"],
            device["serial"],
            net["name"],
            device["model"],
            ",".join(device["tags"]),
            device["firmware"],
        )
    console.print(table)


@app.command()
def update(
    serial: List[str],
    name: Optional[str] = None,
    address: Optional[str] = None,
    notes: Optional[str] = None,
    add_tag: Optional[List[str]] = None,
    remove_tag: Optional[List[str]] = None,
):
    """
    Update parameters of a Meraki device
    """
    if not serial:
        console.print("You must specify a device using `--serial`")
        raise typer.Abort()

    # Get each specified device
    devices = []
    for sn in serial:
        device = dashboard.devices.getDevice(serial=sn)
        console.print(f"Found device named {device.get('name', sn)} with serial {sn}")
        devices.append(device)

    # Confirm same name assignment to multiple devices
    if name and len(devices) > 1:
        console.print(f"[bold red]You specified a name for {len(devices)} devices")
        confirm = Confirm.ask(
            " Do you want to assign the same name to multiple devices?",
            console=console,
        )
        if not confirm:
            raise typer.Abort()

    for device in devices:
        device_sn = device["serial"]
        device_name = device.get("name", device_sn)
        updated_device = {"serial": device_sn}
        # Set new name
        if name:
            console.print(f" Renamed {device_name} to {name}")
            updated_device["name"] = name
            device_name = name

        # Set new address
        if address:
            console.print(f" Assigned address to {device_name}")
            updated_device["address"] = address
            updated_device["moveMapMarker"] = True

        # Set new notes
        if notes:
            console.print(f" Added notes to {device_name}")
            updated_device["notes"] = notes

        if add_tag or remove_tag:
            updated_device["tags"] = device["tags"]
        # Add tags
        if add_tag:
            for tag in add_tag:
                if tag not in updated_device["tags"]:
                    console.print(f" Added tag {tag} to {device_name}")
                    updated_device["tags"].append(tag)

        # Remove tags
        if remove_tag:
            for tag in remove_tag:
                if tag in updated_device["tags"]:
                    console.print(f" Removed tag {tag} to {device_name}")
                    updated_device["tags"].remove(tag)

        device = dashboard.devices.updateDevice(**updated_device)
        console.print(f"Updated device {device_name}")


@app.command()
def show_lldp(
    serial: Optional[List[str]] = None,
    organization_name: Optional[str] = None,
    network_name: Optional[str] = None,
):
    """
    Show CDP/LLDP information for Meraki device(s)
    """
    # Create a list of devices from the specified serial(s)
    devices = []
    for ser in serial:
        device = dashboard.devices.getDevice(serial=ser)
        devices.append(device)

    # If organization name and/or network name are specified, add devices from that scope
    if organization_name:
        if network_name:
            net = find_network_by_name(organization_name, network_name)
            add_devices = dashboard.networks.getNetworkDevices(net["id"])
        else:
            org = find_org_by_name(organization_name)
            add_devices = dashboard.organizations.getOrganizationDevices(org["id"])

        devices += add_devices

    elif network_name:
        console.print("You cannot specify a network name without an organization name.")
        raise typer.Abort()

    if not devices:
        console.print("No devices found.")
        raise typer.Abort()

    console.print(f"Getting CDP/LLDP data for {len(devices)} devices")
    for dev in devices:
        device_lldp = dashboard.devices.getDeviceLldpCdp(dev["serial"])
        if not device_lldp:
            console.print(
                f"No CDP/LLDP data found for {dev.get('name', dev['serial'])}"
            )
            continue

        table = table_with_columns(
            ["Type", "System Name", "Remote Port", "Mgmt Address"],
            title=f"{dev['name']} ({dev['serial']})",
            first_column_name="Port",
        )

        for port, data in device_lldp["ports"].items():
            if port:
                table.add_row(port)
            if "cdp" in data.keys():
                table.add_row(
                    "",
                    "CDP",
                    data["cdp"].get("deviceId"),
                    data["cdp"].get("portId"),
                    data["cdp"].get("address"),
                )
            if "lldp" in data.keys():
                table.add_row(
                    "",
                    "LLDP",
                    data["lldp"].get("systemName"),
                    data["lldp"].get("portId"),
                    data["lldp"].get("managementAddress"),
                )
        console.print(table)


@app.command()
def reboot(serial: List[str] = None):
    """
    Reboot device(s)
    """
    if not serial:
        console.print("No serial numbers entered.")
        raise typer.Abort()

    for sn in serial:
        reboot = dashboard.devices.rebootDevice(serial=sn)
        if reboot["success"]:
            console.print(f"Rebooted device with SN [bold]{sn}")
        else:
            console.print(f"Unabe to reboot device with SN [bold]{sn}")


@app.command()
def blink_led(
    serial: List[str] = None, duration: int = typer.Option(20, min=5, max=120)
):
    """
    Blink the LEDs of device(s)
    """
    if not serial:
        console.print("No serial numbers entered.")
        raise typer.Abort()

    for sn in serial:
        blink = dashboard.devices.blinkDeviceLeds(serial=sn, duration=duration)
        console.print(
            f"Blinking [bold]{sn}[/bold] LEDs for {blink['duration']} seconds"
        )


@app.command()
def ping(
    serial: str,
    target: str = typer.Option(None, help="Specify a target IP or FQDN"),
    count: int = typer.Option(5, min=1, max=5, help="Number of pings"),
):
    """
    Ping a Meraki device
    """
    with status_spinner("Pinging device"):
        # Create a new ping task
        params = {
            "serial": serial,
            "count": count,
        }
        if target is None:
            ping = dashboard.devices.createDeviceLiveToolsPingDevice(**params)
        else:
            params["target"] = target
            ping = dashboard.devices.createDeviceLiveToolsPing(**params)

        # Poll for an update on the ping ttask
        while ping["status"] in ["new", "running"]:
            params = {
                "serial": serial,
                "id": ping["pingId"],
            }
            if target is None:
                ping = dashboard.devices.getDeviceLiveToolsPingDevice(**params)
            else:
                ping = dashboard.devices.getDeviceLiveToolsPing(**params)
            sleep(1)

    # Use 'Meraki cloud' if no target is specified
    if target is None:
        target = "Meraki cloud"

    # Emulate UNIX ping display with statistics
    table = Table(
        show_header=False,
        box=box.HEAVY_EDGE,
        title=f"Ping from Meraki device [bold]{serial}",
    )
    if "replies" in ping["results"].keys():
        for reply in ping["results"]["replies"]:
            table.add_row(
                f"{reply['size']} bytes from {target}:"
                f" icmp_seq={reply['sequenceId']} time={reply['latency']}"
            )
    table.add_row(f"\n--- {target} ping statistics ---")
    table.add_row(
        f"{ping['results']['sent']} packets transmitted,"
        f" {ping['results']['received']} packets received,"
        f" {ping['results']['loss']['percentage']}% packet loss"
    )
    if "latencies" in ping["results"].keys():
        table.add_row(
            "round-trip min/avg/max ="
            f" {ping['results']['latencies']['minimum']}/{ping['results']['latencies']['average']}/{ping['results']['latencies']['maximum']}"
        )
    console.print(table)



"""
merakitools - mr.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
from typing import Optional
from meraki.exceptions import APIError
from rich.prompt import Confirm
import typer
from merakitools.console import console, status_spinner
from merakitools.dashboardapi import dashboard
from merakitools.meraki_helpers import find_network_by_name
from merakitools.formatting_helpers import table_with_columns, camel_case_split
from merakitools.types import (
    DeviceModel,
    MRSSIDIPAssignmentMode,
    ProductType,
    MRSSIDAuthMode,
    MRSSIDEncryptionMode,
    MRSSIDWPAEncrytionMode,
    FirewallPolicyOption,
)

app = typer.Typer()


@app.command()
def list_ssid(
    organization_name: str,
    network_name: str,
    include_disabled: bool = typer.Option(False, "--include-disabled"),
    include_psk: bool = False,
):
    """
    List configured SSIDs for a network
    """
    # Get network and confirm it is wireless
    net = find_network_by_name(organization_name, network_name)
    if ProductType.wireless not in net["productTypes"]:
        console.print(f"This network does not contain any MR devices")
        raise typer.Abort()

    # Get list of SSIDs
    ssids = dashboard.wireless.getNetworkWirelessSsids(net["id"])
    if not include_disabled:
        ssids = [ssid for ssid in ssids if ssid["enabled"]]

    # Create table of SSIDs
    table = table_with_columns(
        [
            "Enabled",
            "Authentication",
            "Mode",
            "VLAN Tag",
            "Band",
            "Visible",
            "Availability",
        ],
        title=f"SSIDs for {net['name']}",
        first_column_name="Name",
    )
    for ssid in ssids:
        table.add_row(
            ssid["name"],
            "[green]Enabled" if ssid["enabled"] else "[red]Disabled",
            f"{ssid.get('encryptionMode', 'open')} / {ssid['authMode']}"
            + f" [bold]{ssid.get('psk') if include_psk else ''}",
            ssid["ipAssignmentMode"],
            str(ssid.get("defaultVlanId")) if ssid.get("useVlanTagging") else "none",
            ssid["bandSelection"],
            "[green]Visible" if ssid["visible"] else "[red]Not visible",
            "All APs"
            if ssid["availableOnAllAps"]
            else f"Tags: {', '.join(ssid['availabilityTags'])}",
        )
    console.print(table)


@app.command()
def list_rf(organization_name: str, network_name: str):
    """
    List RF settings for a network
    """
    # Get network and confirm it is wireless
    net = find_network_by_name(organization_name, network_name)
    if ProductType.wireless not in net["productTypes"]:
        console.print(f"This network does not contain any MR devices")
        raise typer.Abort()

    # Maintain a dict of RF profiles
    rf_profiles = {}

    # Get list of all wireless devices
    with status_spinner("Getting devices"):
        # Get sorted list of MR devices
        devices = dashboard.networks.getNetworkDevices(net["id"])
        devices = [device for device in devices if DeviceModel.MR in device["model"]]
        devices = sorted(devices, key=lambda k: k["name"], reverse=False)

        # Create a table of MR devices with RF info from API
        table = table_with_columns(
            [
                "RF Profile",
                "2.4Ghz Manual Settings",
                "2.4Ghz Actual",
                "5Ghz Manual Settings",
                "5Ghz Actual",
            ],
            title=f"RF Settings for {net['name']}",
            first_column_name="AP Name",
        )
        for device in devices:
            # Get RF settings and current status
            device_rf = dashboard.wireless.getDeviceWirelessRadioSettings(
                serial=device["serial"]
            )
            device_status = dashboard.wireless.getDeviceWirelessStatus(
                serial=device["serial"]
            )

            # Get RF profile from stored data or API
            rf_profile_id = device_rf.get("rfProfileId")
            if rf_profile_id:
                if rf_profile_id not in rf_profiles:
                    rf_profiles[
                        rf_profile_id
                    ] = dashboard.wireless.getNetworkWirelessRfProfile(
                        net["id"], rf_profile_id
                    )

            # Get first SSID on each band for actual status info
            try:
                twoFour_status = next(
                    bss
                    for bss in device_status["basicServiceSets"]
                    if bss["enabled"]
                    and bss["broadcasting"]
                    and bss["band"] == "2.4 GHz"
                )
            except StopIteration:
                # No 2.4Ghz enabled
                twoFour_status = None
            try:
                five_status = next(
                    bss
                    for bss in device_status["basicServiceSets"]
                    if bss["enabled"] and bss["broadcasting"] and bss["band"] == "5 GHz"
                )
            except StopIteration:
                # No 5Ghz enabled
                five_status = None

            # Readable string for 2.4GHz settings
            twoFourGhzSettings = []
            if device_rf["twoFourGhzSettings"]["channel"]:
                twoFourGhzSettings.append(
                    f"ch {device_rf['twoFourGhzSettings']['channel']}"
                )
            if device_rf["twoFourGhzSettings"]["targetPower"] == -1.0:
                twoFourGhzSettings = ["disabled"]
            elif device_rf["twoFourGhzSettings"]["targetPower"]:
                twoFourGhzSettings.append(
                    f"{device_rf['twoFourGhzSettings']['targetPower']} dBm"
                )

            # Readable string for 5GHz settings
            fiveGhzSettings = []
            if device_rf["fiveGhzSettings"]["channel"]:
                fiveGhzSettings.append(f"ch {device_rf['fiveGhzSettings']['channel']}")
            if device_rf["fiveGhzSettings"]["targetPower"] == -1.0:
                fiveGhzSettings = ["disabled"]
            elif device_rf["fiveGhzSettings"]["targetPower"]:
                fiveGhzSettings.append(
                    f"{device_rf['fiveGhzSettings']['targetPower']} dBm"
                )

            table.add_row(
                device["name"],
                rf_profiles[rf_profile_id]["name"] if rf_profile_id else "None",
                " / ".join(twoFourGhzSettings),
                f"ch {twoFour_status['channel']} / {twoFour_status['power']}"
                if twoFour_status
                else "Not broadcasting",
                " / ".join(fiveGhzSettings),
                f"ch {five_status['channel']} / {five_status['power']}"
                if five_status
                else "Not broadcasting",
            )
    console.print(table)


@app.command()
def list_rf_profiles(organization_name: str, network_name: str):
    """
    List RF profiles for a network
    """
    # Get network and confirm it is wireless
    net = find_network_by_name(organization_name, network_name)
    if ProductType.wireless not in net["productTypes"]:
        console.print(f"This network does not contain any MR devices")
        raise typer.Abort()

    # Get RF profiles
    with status_spinner("Getting RF profiles"):
        rf_profiles = dashboard.wireless.getNetworkWirelessRfProfiles(
            net["id"], includeTemplateProfiles=True
        )
        if not rf_profiles:
            console.print("No RF Profiles found")
            raise typer.Abort()

        table = table_with_columns(
            [
                "Client Balancing",
                "Band Settings",
                "2.4GHz Power",
                "2.4GHz Min. Bitrate",
                "2.4GHz AutoChannel",
                "5GHz Power",
                "5GHz Min. Bitrate",
                "5GHz AutoChannel",
                "5GHz Width",
            ],
            title=f"RF Profiles for {net['name']}",
            first_column_name="Name",
        )

        for profile in rf_profiles:
            twoFourGhzChannels = [
                str(int) for int in profile["twoFourGhzSettings"]["validAutoChannels"]
            ]
            fiveGhzChannels = [
                str(int) for int in profile["fiveGhzSettings"]["validAutoChannels"]
            ]

            table.add_row(
                profile["name"],
                "[green]Enabled"
                if profile["clientBalancingEnabled"]
                else "[red]Disabled",
                f"{profile['apBandSettings']['bandOperationMode']} /"
                f" {'[green]Band Steering Enabled' if profile['apBandSettings']['bandSteeringEnabled'] else '[red]Band Steering Disabled'}"
                if profile["bandSelectionType"] == "ap"
                else "per SSID",
                f"{profile['twoFourGhzSettings']['minPower']}-{profile['twoFourGhzSettings']['maxPower']}dBm",
                f"{profile['twoFourGhzSettings']['minBitrate']}Mbps"
                if profile["minBitrateType"] == "band"
                else "per SSID",
                ", ".join(twoFourGhzChannels),
                f"{profile['fiveGhzSettings']['minPower']}-{profile['fiveGhzSettings']['maxPower']}dBm",
                f"{profile['fiveGhzSettings']['minBitrate']}Mbps"
                if profile["minBitrateType"] == "band"
                else "per SSID",
                ", ".join(fiveGhzChannels),
                profile["fiveGhzSettings"]["channelWidth"]
                if profile["fiveGhzSettings"]["channelWidth"] == "auto"
                else f"{profile['fiveGhzSettings']['channelWidth']}MHz",
                end_section=True,
            )
        console.print(table)


@app.command()
def list_mesh(organization_name: str, network_name: str):
    """
    List mesh status for a network
    """
    # Get network and confirm it is wireless
    net = find_network_by_name(organization_name, network_name)
    if ProductType.wireless not in net["productTypes"]:
        console.print(f"This network does not contain any MR devices")
        raise typer.Abort()

    # Get network mesh status
    with status_spinner("Getting mesh status"):
        try:
            mesh = dashboard.wireless.getNetworkWirelessMeshStatuses(net["id"])
        except APIError as err:
            console.print(f"[bold]{err.message['errors'][0]}")
            raise typer.Abort()

    # Maintain a dict of mesh device info
    devices = {}

    # Create a table of mesh devices and routes
    table = table_with_columns(
        ["Mesh Route", "Mbps", "Metric", "Usage"],
        title=f"Mesh Status for {net['name']}",
        first_column_name="AP Name",
    )
    for ap in mesh:
        mesh_route = ""
        # Iterate through each item in the mesh route
        for idx, serial in enumerate(ap["meshRoute"]):
            # Get device data for new serials
            if serial not in devices:
                devices[serial] = dashboard.devices.getDevice(serial)

            # Add mesh name to mesh route
            last_item = idx == len(ap["meshRoute"]) - 1
            mesh_route += f"[bold]{devices[serial]['name']}[/bold]"
            if not last_item:
                mesh_route += " -> "

        table.add_row(
            devices[ap["serial"]]["name"],
            mesh_route,
            f"{ap['latestMeshPerformance']['mbps']}Mbps",
            str(ap["latestMeshPerformance"]["metric"]),
            ap["latestMeshPerformance"]["usagePercentage"],
        )
    console.print(table)


@app.command()
def show_ssid(
    organization_name: str,
    network_name: str,
    ssid_number: int = typer.Argument(..., min=0, max=15),
):
    """
    Show an SSID for a network

    TODO: formatting
    """
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting SSID"):
        ssid = dashboard.wireless.getNetworkWirelessSsid(
            networkId=net["id"], number=ssid_number
        )

    console.print(ssid)


@app.command()
def update_ssid(
    organization_name: str,
    network_name: str,
    ssid_number: int = typer.Argument(..., help="The SSID number", min=0, max=15),
    confirm: bool = typer.Option(
        True, help="Confirm the current SSID name before applying changes"
    ),
    enabled: Optional[bool] = typer.Option(
        None, help="Whether or not the SSID is enabled"
    ),
    name: Optional[str] = typer.Option(None, help="The name of the SSID"),
    auth_mode: Optional[MRSSIDAuthMode] = typer.Option(
        None, help="The association control method for the SSID"
    ),
    encryption_mode: Optional[MRSSIDEncryptionMode] = typer.Option(
        None, help="The PSK encryption mode for the SSID"
    ),
    wpa_encryption_mode: Optional[MRSSIDWPAEncrytionMode] = typer.Option(
        None, help="The types of WPA encryption"
    ),
    tag_vlan: Optional[bool] = typer.Option(
        None, help="Whether or not traffic shuold be directed to use specific VLANs"
    ),
    default_vlan_id: Optional[int] = typer.Option(
        None, help="The default VLAN ID used for 'all other APs'", min=1, max=4094
    ),
    pre_shared_key: Optional[str] = typer.Option(None, help="The passkey for the SSID"),
    min_bitrate: Optional[int] = typer.Option(
        None, help="The minimum bitrate in Mbps", min=1, max=54
    ),
    ip_assignment_mode: Optional[MRSSIDIPAssignmentMode] = typer.Option(
        None, help="The client IP assignment mode"
    ),
    local_lan_access: Optional[FirewallPolicyOption] = typer.Option(
        None, help="Policy for wireless clients accessing the Local LAN"
    ),
):
    """
    Update an SSID for a network
    """
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting SSIDs"):
        ssid = dashboard.wireless.getNetworkWirelessSsid(
            networkId=net["id"], number=ssid_number
        )
        local_lan_rule = (
            dashboard.wireless.getNetworkWirelessSsidFirewallL3FirewallRules(
                networkId=net["id"],
                number=ssid_number,
            )["rules"][0]
        )

    # Confirm SSID name with user before continuing
    if confirm:
        console.print(
            f"SSID number [bold]{ssid_number}[/bold] is named"
            f" [bold]{ssid['name']}[/bold]"
        )
        confirmed = Confirm.ask("Do you want to continue?", console=console)
        if not confirmed:
            raise typer.Abort()

    # Enumerate changed values for SSID
    update_ssid = {}
    items = {
        "name": name,
        "enabled": enabled,
        "useVlanTagging": tag_vlan,
        "defaultVlanId": default_vlan_id,
        "authMode": auth_mode.value if auth_mode is not None else None,
        "encryptionMode": encryption_mode.value
        if encryption_mode is not None
        else None,
        "wpaEncryptionMode": wpa_encryption_mode.value
        if wpa_encryption_mode is not None
        else None,
        "minBitrate": min_bitrate,
        "psk": pre_shared_key,
        "ipAssignmentMode": ip_assignment_mode.value
        if ip_assignment_mode is not None
        else None,
    }
    for key, value in items.items():
        if value is not None:
            if value != ssid.get(key, None):
                update_ssid[key] = value

    # Enumerate changed value for Local LAN FW
    update_lan_fw = {}
    if local_lan_access is not None:
        if local_lan_access.value != local_lan_rule.get("policy"):
            update_lan_fw["allowLanAccess"] = (
                True if local_lan_access.value == "allow" else False
            )

    # Do not call API if no changes were made
    with status_spinner("Updating SSID"):
        changed = False
        if update_ssid:
            changed = True
            ssid = dashboard.wireless.updateNetworkWirelessSsid(
                networkId=net["id"], number=ssid_number, **update_ssid
            )

        if update_lan_fw:
            changed = True
            local_lan_rule = (
                dashboard.wireless.updateNetworkWirelessSsidFirewallL3FirewallRules(
                    networkId=net["id"], number=ssid_number, **update_lan_fw
                )
            )
    if not changed:
        console.print(f"[bold green]No settings changed.")
        raise typer.Exit()

    console.print(f"[bold green]SSID '{ssid['name']}' has been updated.")
    console.print(
        " The following parameters were updated:"
        f" {', '.join(update_ssid | update_lan_fw)}"
    )


@app.command()
def list_l3_fw(
    organization_name: str,
    network_name: str,
    ssid_number: int = typer.Argument(..., help="The SSID number", min=0, max=15),
):
    """
    List layer 3 firewall rules for an SSID
    """
    net = find_network_by_name(organization_name, network_name)
    # Get SSID details and rules
    with status_spinner("Getting SSIDs"):
        ssid = dashboard.wireless.getNetworkWirelessSsid(
            networkId=net["id"],
            number=ssid_number,
        )
        rules = dashboard.wireless.getNetworkWirelessSsidFirewallL3FirewallRules(
            networkId=net["id"],
            number=ssid_number,
        )["rules"]

    # Print SSID name and rules in table
    console.print(f"SSID: [bold]{ssid['name']}")
    table = table_with_columns(
        ["Policy", "Protocol", "Dest IP", "Dest Port", "Comment"],
        title=f"L3 rules for SSID: [bold]{ssid['name']}",
    )
    for rule in rules:
        table.add_row(
            rule["policy"],
            rule["protocol"],
            rule["destCidr"],
            rule["destPort"],
            rule["comment"],
        )
    console.print(table)


@app.command()
def list_l7_fw(
    organization_name: str,
    network_name: str,
    ssid_number: int = typer.Argument(..., help="The SSID number", min=0, max=15),
):
    """
    List layer 7 firewall rules for an SSID
    """
    net = find_network_by_name(organization_name, network_name)
    # Get SSID details and rules
    with status_spinner("Getting SSIDs"):
        ssid = dashboard.wireless.getNetworkWirelessSsid(
            networkId=net["id"],
            number=ssid_number,
        )
        rules = dashboard.wireless.getNetworkWirelessSsidFirewallL7FirewallRules(
            networkId=net["id"],
            number=ssid_number,
        )["rules"]

    # Print SSID name and rules in table
    console.print(f"SSID: [bold]{ssid['name']}")
    if not rules:
        console.print("No rules found.")
        raise typer.Exit()

    table = table_with_columns(
        ["Policy", "Type", "Name"], title=f"L7 rules for SSID: [bold]{ssid['name']}"
    )
    for rule in rules:
        table.add_row(
            rule["policy"],
            camel_case_split(rule["type"]),
            rule["value"] if type(rule["value"]) is str else rule["value"]["name"],
        )
    console.print(table)



"""
merakitools - mt.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""

from typing import List, Optional
import typer
from merakitools.meraki_helpers import find_org_by_name
from merakitools.types import MTMetricType
from merakitools.console import console
from merakitools.dashboardapi import dashboard
from merakitools.formatting_helpers import table_with_columns

app = typer.Typer()


@app.command()
def latest_readings(
    organization_name: str,
    serial: Optional[List[str]] = None,
):
    """
    Show the latest reading for each metric from each sensor
    """
    org = find_org_by_name(organization_name)
    readings = dashboard.sensor.getOrganizationSensorReadingsLatest(org["id"])

    for device in readings:
        if serial and device["serial"] not in serial:
            continue

        table = table_with_columns(
            ["Data", "Time"],
            title=f"{device['network']['name']}: {device['serial']}",
            first_column_name="Metric",
        )
        for reading in device["readings"]:
            table.add_row(
                reading["metric"].capitalize(),
                str(reading[reading["metric"]]),
                reading["ts"],
            )

        console.print(table)


@app.command()
def history(
    organization_name: str,
    serial: Optional[List[str]] = None,
    metric_type: Optional[List[MTMetricType]] = None,
):
    """
    Show the historical sensor readings for an organization
    """
    org = find_org_by_name(organization_name)
    readings = dashboard.sensor.getOrganizationSensorReadingsHistory(org["id"])

    table = table_with_columns(
        ["Data", "Time", "Network / Serial"], "History", first_column_name="Metric"
    )
    for reading in readings:
        if metric_type and reading["metric"] not in metric_type:
            continue

        if serial and reading["serial"] not in serial:
            continue

        table.add_row(
            reading["metric"],
            str(reading[reading["metric"]]),
            reading["ts"],
            f"{reading['network']['name']} / {reading['serial']}",
        )
    console.print(table)



"""
merakitools - main.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
import sys
import typer
from merakitools import orgs, networks, devices, mx, ms, mr, mt, msp

# Python 3.9+ is required
MIN_PYTHON = (3, 9)
if sys.version_info < MIN_PYTHON:
    sys.exit(f"Python {MIN_PYTHON[0]}.{MIN_PYTHON[1]} or later is required")

app = typer.Typer()
app.add_typer(orgs.app, name="orgs", help="Meraki organizations")
app.add_typer(networks.app, name="networks", help="Meraki networks")
app.add_typer(devices.app, name="devices", help="Meraki devices")
app.add_typer(mx.app, name="mx", help="Meraki MX appliances")
app.add_typer(ms.app, name="ms", help="Meraki MS switches")
app.add_typer(mr.app, name="mr", help="Meraki MR wireless")
app.add_typer(mt.app, name="mt", help="Meraki MT sensors")
app.add_typer(msp.app, name="msp", help="Manage multiple networks")



"""
merakitools - console.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
from rich.console import Console
from rich.traceback import install

console = Console()
install(show_locals=True, console=console)


def status_spinner(message: str):
    """
    Common spinner
    """
    return console.status(f"{message}..", spinner="material")



"""
merakitools - meraki_helpers.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
import os
from typing import Optional, List
from meraki.exceptions import APIError
import requests
import typer
from merakitools.console import console
from merakitools.dashboardapi import dashboard


def find_orgs_by_name(org_name: Optional[str]) -> List:
    """
    Given a name, find any matching organizations
    """
    with console.status("Finding Organizations..", spinner="material"):
        orgs = dashboard.organizations.getOrganizations()
    if org_name:
        orgs = [org for org in orgs if org_name in org["name"]]
    return orgs


def find_org_by_name(org_name: str):
    """
    Accepts an organization name or ID, and return the Meraki organization
    """
    with console.status("Finding organization..", spinner="material"):
        # Use numeric values as Org IDs
        if org_name.isnumeric():
            try:
                # Get org by ID
                org = dashboard.organizations.getOrganization(organizationId=org_name)
            except APIError as exc:
                console.print(
                    f"Organization ID [bold]{org_name}[/bold] not accessible."
                )
                raise typer.Abort() from exc
        else:
            try:
                # Get list of accessible orgs and search by name
                orgs = dashboard.organizations.getOrganizations()
                org = next(org for org in orgs if org["name"] == org_name)
            except APIError as err:
                console.print(f"{err.message}")
                raise typer.Abort()
            except StopIteration as exc:
                console.print(f"Organization named [bold]{org_name}[/bold] not found.")
                raise typer.Abort() from exc

    console.print(f"Organization: [bold]{org['name']}")
    return org


def find_org_id_by_device_serial(serial: str):
    """
    Given a serial, find the organization it belongs to
    """
    try:
        device = dashboard.devices.getDevice(serial=serial)
    except APIError as exc:
        console.print(f"[red]Unable to find device with serial '{serial}'")
        raise typer.Abort() from exc

    network = dashboard.networks.getNetwork(networkId=device["networkId"])

    return network["organizationId"]


def find_network_by_name(org_name: str, net_name: str):
    """
    Find a network given an orgaization name and network name
    """
    org = find_org_by_name(org_name)
    with console.status("Finding network..", spinner="material"):
        nets = dashboard.organizations.getOrganizationNetworks(org["id"])

    try:
        net = next(net for net in nets if net["name"] == net_name)
    except StopIteration as exc:
        print("Network not found.")
        raise typer.Abort() from exc

    console.print(f"Network: [bold]{net['name']}")
    return net


def api_req(resource: str, method: str = "GET", **kwargs):
    """
    API request outside of the Meraki Python SDK
    """
    base_url = "https://api.meraki.com/api/v1"
    headers = {
        "Accept": "application/json",
        "X-Cisco-Meraki-API-Key": os.getenv("MERAKI_DASHBOARD_API_KEY"),
    }

    resp = requests.request(
        url=f"{base_url}/{resource}", method=method, headers=headers, **kwargs
    )

    resp.raise_for_status()
    if resp.text:
        return resp.json()

    return {}



"""
merakitools - networks.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
from enum import Enum


class ProductType(str, Enum):
    """
    Types of Meraki products
    """

    appliance = "appliance"
    switch = "switch"
    wireless = "wireless"
    camera = "camera"
    systemsManager = "systemsManager"
    enviornmental = "enviornmental"
    sensor = "sensor"
    cellularGateway = "cellularGateway"


class DeviceModel(str, Enum):
    """
    Models of Meraki devices
    """

    MX = "MX"
    MR = "MR"
    MS = "MS"
    MV = "MV"
    MT = "MT"
    Z = "Z"


class DeviceSortOptions(str, Enum):
    """
    Options to sort a list of devices
    """

    name = "name"
    model = "model"


class FirewallPolicyOption(str, Enum):
    """
    Firewall policy options
    """

    allow = "allow"
    deny = "deny"


class MXInternetUplinks(str, Enum):
    """
    Uplink types on an MX device
    """

    one = "internet1"
    two = "internet2"


class TrafficDirection(str, Enum):
    """
    Directions traffic can flow from an interface
    """

    total = "total"
    recv = "recv"
    sent = "sent"


class MSInterfaceMode(str, Enum):
    """
    Interface modes on an MS switch
    """

    access = "access"
    trunk = "trunk"


class MSSTPGuardType(str, Enum):
    """ "
    STP Guard modues on an MS switch
    """

    disabled = "disabled"
    root = "root guard"
    bpdu = "bpdu guard"
    loop = "loop guard"


class MRSSIDAuthMode(str, Enum):
    """
    Authentication modes on an MR SSID
    """

    open = "open"
    psk = "psk"
    open_with_radius = "open-with-radius"
    dot1x_meraki = "8021x-meraki"
    dot1x_radius = "8021x-radius"
    dot1x_google = "8021x-google"
    dot1x_localradius = "8021x-localradius"
    ipsk_with_radius = "ipsk-with-radius"
    ipsk_without_radius = "ipsk-without-radius"


class MRSSIDEncryptionMode(str, Enum):
    """
    Encryption modes on an MR SSID
    """

    wep = "wep"
    wpa = "wpa"


class MRSSIDWPAEncrytionMode(str, Enum):
    """
    WPA Encryption modes on an MR SSID
    """

    wpa1_only = "WPA1 only"
    wpa12 = "WPA1 and WPA2"
    wpa2_only = "WPA2 only"
    wpa3_transition = "WPA3 Transition Mode"
    wpa3_only = "WPA3 only"


class MRSSIDIPAssignmentMode(str, Enum):
    """
    IP assignment mode on an MR SSID
    """

    nat_mode = "NAT mode"
    bridge_mode = "Bridge mode"
    l3_roaming = "Layer 3 roaming"
    l3_roaming_concentrator = "Layer 3 roaming with a concentrator"
    vpn = "VPN"


class NetworkTrafficAnalysisMode(str, Enum):
    """
    Traffic Analysis mode for a network
    """

    disabled = "disabled"
    basic = "basic"
    detailed = "detailed"


class MTMetricType(str, Enum):
    """
    Meraki MT metric types
    """

    battery = "battery"
    temperature = "temperature"
    humidity = "humidity"
    water = "water"
    door = "door"
    noise = "noise"
    co2 = "co2"
    tvoc = "tvoc"
    indoorAirQuality = "indoorAirQuality"
    pm25 = "pm25"
    button = "button"



"""
merakitools - msp.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""

from typing import List, Optional
from datetime import datetime, timedelta
import typer
from rich.progress import track
from merakitools.console import console, status_spinner
from merakitools.dashboardapi import dashboard, APIError


app = typer.Typer()


@app.command()
def enable_api_all():
    """
    Enable the Meraki API for all accessible organizations
    """
    with status_spinner("Finding organizations"):
        orgs = dashboard.organizations.getOrganizations()
        orgs = [org for org in orgs if not org["api"]["enabled"]]
    console.print(f"Found {len(orgs)} organizations with API disabled.")

    for org in track(orgs, console=console):
        try:
            org = dashboard.organizations.updateOrganization(
                organizationId=org["id"], name=org["name"], api={"enabled": True}
            )
        except APIError:
            console.print(f"[red]Error enabling API for {org['name']}")
            continue
        console.print(f"[green]API enabled for {org['name']}")


@app.command()
def list_security_events(
    days_ago: int = typer.Option(3, help="How many days to look back for events"),
    organization_name: Optional[List[str]] = typer.Option(
        None, help="Specify organization"
    ),
    filter_event_name: Optional[str] = typer.Option(
        None, help="Filter by event message"
    ),
):
    """
    List security events for organization(s), filtering by organization or event name
    """
    # Get all accessible organizations, filtering by name if specified
    with status_spinner("Finding organizations"):
        orgs = dashboard.organizations.getOrganizations()
        if organization_name:
            orgs = [org for org in orgs if org["name"] in organization_name]
    console.print(f"[bold]Found {len(orgs)} organizations.")

    # Calculate number of days to look back
    end_time = datetime.now()
    start_time = end_time - timedelta(days=days_ago)
    start_time = datetime.combine(start_time, start_time.min.time())
    dtformat = "%Y-%m-%d %H:%M:%S"
    console.print(
        f" [italic]Displaying events from {start_time.strftime(dtformat)} to"
        f" {end_time.strftime(dtformat)}."
    )

    # Iterate through each org to gather security events
    for org in track(orgs, console=console):
        try:
            events = dashboard.appliance.getOrganizationApplianceSecurityEvents(
                organizationId=org["id"],
                total_pages="all",
                perPage=1000,
                t0=start_time,
                t1=end_time,
            )
        except APIError:
            continue

        # Iterate through each event
        event_hosts = []
        filtered_events = []
        for event in events:
            # Drop events that do not match the filter
            if filter_event_name is not None:
                if not filter_event_name in event["message"]:
                    continue
            filtered_events.append(event)

            # Maintain a list of associated hosts
            if not event["destIp"] in event_hosts:
                event_hosts.append(event["destIp"])

        if len(filtered_events) > 0:
            console.print(
                f"[bold]{org['name']}[/bold] matched {len(filtered_events)} events."
            )
            for host in event_hosts:
                console.print(f" {host}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['merakitools']

package_data = \
{'': ['*']}

install_requires = \
['meraki>=1.22.1,<2.0.0', 'typer[all]>=0.5.0,<0.6.0']

entry_points = \
{'console_scripts': ['merakitools = merakitools.main:app']}

setup_kwargs = {
    'name': 'merakitools',
    'version': '0.1.10',
    'description': 'CLI tools for managing Meraki networks based on Typer',
    'long_description': '# merakitools\n[![published](https://static.production.devnetcloud.com/codeexchange/assets/images/devnet-published.svg)](https://developer.cisco.com/codeexchange/github/repo/billyzoellers/merakitools)\n[![CI](https://github.com/billyzoellers/merakitools/actions/workflows/ci.yml/badge.svg?branch=main)](https://github.com/billyzoellers/merakitools/actions/workflows/ci.yml)\n[![linting: pylint](https://img.shields.io/badge/linting-pylint-yellowgreen)](https://github.com/PyCQA/pylint)\n[![security: bandit](https://img.shields.io/badge/security-bandit-yellow.svg)](https://github.com/PyCQA/bandit)\n\nCLI tools for managing Meraki networks based on [Typer](https://typer.tiangolo.com/).\n\nmerakitools provides CLI based to monitoring and configuration tools available through the Meraki\nDashboard API. It was designed to help with bulk configuration creation and migrations to the Meraki platform.\n\n## Installation\nInstall with `pip` or your favorite PyPi package mananger.\n```\npip install merakitools\n```\n\nSet enviornment variable `MERAKI_DASHBOARD_API_KEY` to your Meraki API key\n\nThen try a command.\n```\nmerakitools orgs list\n```\n\n## Example Commands\nList Meraki Networks in an organization\n```\nmerakitools networks list <YourOrgName>\n```\n\nList Meraki MR devices in a network\n```\nmerakitools devices list <YourOrgName> <YourNetworkName> --type MR\\\n```\n\nCreate a static NAT entry on a Meraki MX security appliance\n```\nmerakitools mx create-staticnat <YourOrgName> <YourNetworkName> --nat <name>!<publicIP>!<privateIP> --port tcp!636!192.0.2.1/32 --port tcp!8080!any\n```\n\n***For more commands check out the [command documentation](COMMANDS.md).***\n\n## Testing\nFor a free and easy to use testing enviornment, use the Cisco DevNet Sandbox [Meraki](https://developer.cisco.com/docs/sandbox/#!networking/meraki).\n*Note: The sandbox is read-only, so you will not be able to test commands that write data to the Dashboard*\n\n## License\nCopyright (C) 2021  Billy Zoellers\n\nThis program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.\nThis program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.\nYou should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.\n',
    'author': 'Billy Zoellers',
    'author_email': 'billy.zoellers@mac.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""
merakitools - networks.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
from typing import Optional
from rich.prompt import Confirm
import typer
from merakitools.console import console, status_spinner
from merakitools.dashboardapi import dashboard
from merakitools.meraki_helpers import (
    find_org_by_name,
    find_network_by_name,
    api_req,
)
from merakitools.formatting_helpers import table_with_columns, table_network_health
from merakitools.types import ProductType, NetworkTrafficAnalysisMode

app = typer.Typer()


@app.command()
def list(organization_name: str, product_type: Optional[ProductType] = None):
    """
    List Meraki networks in an organization
    """
    # Find an organization and get associated networks
    org = find_org_by_name(organization_name)
    with status_spinner("Getting networks"):
        networks = dashboard.organizations.getOrganizationNetworks(org["id"])

    # Filter based on input and sort alphabetically
    if product_type:
        networks = [
            net for net in networks if product_type.value in net["productTypes"]
        ]
    networks = sorted(networks, key=lambda i: i["name"])

    # Display a table
    table = table_with_columns(["Name", "Type", "ID", "Time Zone"])
    for net in networks:
        table.add_row(
            net["name"], ", ".join(net["productTypes"]), net["id"], net["timeZone"]
        )
    console.print(table)


@app.command()
def update_settings(
    organization_name: str,
    network_name: str,
    confirm: bool = typer.Option(
        True, help="Confirm the network name before applying changes"
    ),
    local_status: bool = typer.Option(
        None,
        "--enable-local-status/--disable-local-status",
        help="Local device status pages",
    ),
    remote_status: bool = typer.Option(
        None,
        "--enable-remote-status/--disable-remote-status",
        help="Remote device status pages",
    ),
):
    """
    Update network settings
    """
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting current settings"):
        settings = dashboard.networks.getNetworkSettings(networkId=net["id"])

    # Confirm SSID name with user before continuing
    if confirm:
        console.print(f"Network is named [bold]{net['name']}[/bold]")
        confirmed = Confirm.ask("Do you want to continue?", console=console)
        if not confirmed:
            raise typer.Abort()

    update = {}
    items = {
        "localStatusPageEnabled": local_status,
        "remoteStatusPageEnabled": remote_status,
    }
    for key, value in items.items():
        if value is not None:
            if value is not settings.get(key, None):
                update[key] = value

    # Do not call API if no changes were made
    if not update:
        console.print("[bold green]No settings changed.")
        raise typer.Exit()

    # Update settings
    with status_spinner("Updating settings"):
        settings = dashboard.networks.updateNetworkSettings(
            networkId=net["id"], **update
        )
    console.print(f"[bold green]Settings for'{net['name']}' have been updated.")
    console.print(f" The following parameters were updated: {', '.join(update)}")


@app.command()
def health(organization_name: str, network_name: str):
    """
    Get Meraki network health (global alerts) for a network
    """
    # Find a network and get network health alerts
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Gathering health information"):
        health = dashboard.networks.getNetworkHealthAlerts(net["id"])

    # Print number of alerts found, exist it no alerts
    if not health:
        console.print(f"[green]No health alerts found for {net['name']}!")
        raise typer.Exit()
    console.print(f"Found {len(health)} alerts for {net['name']}")

    console.print(
        table_network_health(health, title=f"Network health for {net['name']}")
    )


@app.command()
def traffic_analysis(
    organization_name: str,
    network_name: str,
    confirm: bool = typer.Option(
        True, help="Confirm the network name before applying changes"
    ),
    set_mode: NetworkTrafficAnalysisMode = typer.Option(
        None, help="Traffic analysis mode for network"
    ),
):
    """
    Get or update the traffic analysis mode for a network
    """
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting current settings"):
        traffic_analysis = dashboard.networks.getNetworkTrafficAnalysis(
            networkId=net["id"]
        )

    # Print current mode
    console.print(f"Current Traffic analysis mode: [bold]{traffic_analysis['mode']}")

    # Update mode if requested
    if set_mode is not None:
        # Confirm network name with user before continuing
        if confirm:
            confirmed = Confirm.ask("Do you want to continue?", console=console)
            if not confirmed:
                raise typer.Abort()

        # Make change only if required
        if set_mode.lower() == traffic_analysis["mode"].lower():
            console.print("[bold green]No settings changed.")
            raise typer.Abort()

        # Update settings
        with status_spinner("Updating settings"):
            dashboard.networks.updateNetworkTrafficAnalysis(
                networkId=net["id"],
                mode=set_mode,
            )
        console.print(
            f"[bold green]Traffic analysis mode for '{net['name']}' changed to"
            f" {set_mode}"
        )


@app.command()
def list_firmware_upgrades(
    organization_name: str,
    network_name: str,
):
    """
    List firmware upgrades for a network
    """
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting firmware information"):
        firmware_upgrades = dashboard.networks.getNetworkFirmwareUpgrades(
            networkId=net["id"]
        )

    table = table_with_columns(
        [
            "Current Version",
            "Last Upgrade",
            "Last Upgrade Time",
            "Available Versions",
            "Upgrade Scheduled",
            "Upgrade Time",
        ],
        first_column_name="Product",
    )
    for product, fw_info in firmware_upgrades["products"].items():
        available_versions = [
            av["shortName"]
            for av in fw_info["availableVersions"]
            if av["id"] != fw_info["currentVersion"]["id"]
        ]

        table.add_row(
            product.capitalize(),
            f"[bold]{fw_info['currentVersion']['shortName']}[/bold]"
            f" ({fw_info['currentVersion']['releaseType']})",
            f"[bold]{fw_info['lastUpgrade']['fromVersion'].get('shortName', None)}[/bold]"
            f" -> [bold]{fw_info['lastUpgrade']['toVersion'].get('shortName', None)}[/bold]",
            fw_info["lastUpgrade"]["time"],
            ", ".join(available_versions),
            f"{fw_info['nextUpgrade']['toVersion']['shortName']} ({fw_info['nextUpgrade']['toVersion']['releaseType']})"
            if fw_info["nextUpgrade"]["time"]
            else "",
            fw_info["nextUpgrade"]["time"] if fw_info["nextUpgrade"]["time"] else "",
        )

    console.print(table)


@app.command()
def list_webhook_servers(
    organization_name: str,
    network_name: str,
):
    """
    List webhook servers for a network
    """
    # Get a list of the current webhook servers
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting current settings"):
        http_servers = dashboard.networks.getNetworkWebhooksHttpServers(
            networkId=net["id"]
        )

    # Output to table
    table = table_with_columns(
        ["URL"], first_column_name="Name", title=f"Webhook Servers for {net['name']}"
    )
    for server in http_servers:
        table.add_row(server["name"], server["url"])
    console.print(table)


@app.command()
def new_webhook_server(
    organization_name: str,
    network_name: str,
    name: str = typer.Option(..., help="A name for easy reference"),
    shared_secret: str = typer.Option(
        None, help="A shared secret included in POSTs send to the HTTP server"
    ),
    url: str = typer.Option(
        ..., help="The URL of the HTTP server. Cannot be updated later."
    ),
):
    """
    Create a new webhook server for a network
    """
    # Get a list of the current webhook servers
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting current webhook servers"):
        http_servers = dashboard.networks.getNetworkWebhooksHttpServers(
            networkId=net["id"]
        )

    # Validate inputs
    duplicate_name = duplicate_url = None
    for server in http_servers:
        if server["name"] == name:
            duplicate_name = server
            break
        if server["url"] == url:
            duplicate_url = server
    if duplicate_name:
        console.print(f"[red]A webhook server named {name} already exists!")
        raise typer.Abort()
    if duplicate_url:
        console.print(
            f"[red]A webhook server named {duplicate_url['name']} already exists with"
            " this URL!"
        )
        raise typer.Abort()

    # Create new webhook server
    new_server = dashboard.networks.createNetworkWebhooksHttpServer(
        networkId=net["id"],
        name=name,
        url=url,
        sharedSecret=shared_secret if shared_secret else "",
    )
    console.print(
        f"[bold green]New webhook server named '{new_server['name']}' created."
    )


@app.command()
def list_payload_templates(
    organization_name: str,
    network_name: str,
):
    """
    List webhook payload templates for a network
    """
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting templates"):
        templates = api_req(f"networks/{net['id']}/webhooks/payloadTemplates")

    # Output to table
    table = table_with_columns(
        ["Type", "ID"],
        first_column_name="Name",
        title=f"Payload templates for {net['name']}",
    )
    for template in templates:
        table.add_row(
            template["name"],
            template["type"],
            template["payloadTemplateId"],
        )
    console.print(table)


@app.command()
def new_payload_template(
    organization_name: str,
    network_name: str,
    name: str = typer.Option(..., help="A name for the payload template"),
    headers: typer.FileBinaryRead = typer.Option(
        ..., help="A file with the headers template"
    ),
    body: typer.FileBinaryRead = typer.Option(
        ..., help="A file with the body template"
    ),
):
    """
    Create a webhook payload template
    """
    net = find_network_by_name(organization_name, network_name)

    with status_spinner(f"Creating template [bold]{name}"):
        data = {"name": name}
        files = {
            "headersFile": headers,
            "bodyFile": body,
        }

        template = api_req(
            f"networks/{net['id']}/webhooks/payloadTemplates",
            method="POST",
            data=data,
            files=files,
        )

    console.print(
        f"[green]Created new template named [bold]{template['name']}[/bold] with ID"
        f" {template['payloadTemplateId']}."
    )


@app.command()
def delete_payload_template(
    organization_name: str,
    network_name: str,
    name: str = typer.Option(..., help="Name of the payload template"),
    confirm: bool = typer.Option(True, help="Confirm before deleting"),
):
    """
    Delete a webhook payload template
    """
    # Get a list of payload templates for the network
    net = find_network_by_name(organization_name, network_name)
    with status_spinner("Getting templates"):
        templates = api_req(f"networks/{net['id']}/webhooks/payloadTemplates")

    # Search by name
    try:
        found = next(t for t in templates if t["name"] == name)
    except StopIteration as exc:
        console.print(f"[red]Payload template named [bold]{name}[/bold] not found.")
        raise typer.Abort() from exc
    console.print(f"Found template named [bold]{found['name']}")

    # Confirm with user
    if confirm:
        confirmed = Confirm.ask("Do you want to delete?", console=console)
        if not confirmed:
            raise typer.Abort()

    api_req(
        f"networks/{net['id']}/webhooks/payloadTemplates/{found['payloadTemplateId']}",
        method="DELETE",
    )
    console.print(f"[green]Deleted payload template [bold]{name}")



"""
merakitools - dashboard.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
import meraki
from meraki.exceptions import APIError
from merakitools.__init__ import __version__
from merakitools.console import console

try:
    dashboard = meraki.DashboardAPI(
        output_log=False,
        print_console=False,
        suppress_logging=True,
        caller=f"merakitools/{__version__}",
    )
except APIError:
    console.print("[bold red]Unable to connect to the Meraki Dashboard.")
    console.print_exception()



"""
merakitools - orgs.py
Billy Zoellers

CLI tools for managing Meraki networks based on Typer
"""
from typing import List, Optional
import typer
from merakitools.console import console, status_spinner
from merakitools.dashboardapi import dashboard, APIError
from merakitools.meraki_helpers import (
    find_network_by_name,
    find_org_by_name,
    find_orgs_by_name,
    api_req,
)
from merakitools.formatting_helpers import table_with_columns, table_network_health

app = typer.Typer()


@app.command()
def list(name: Optional[str] = None, include_counts: bool = False):
    """
    List Meraki organizations
    """
    orgs = find_orgs_by_name(name)
    console.print(f"Found [bold]{len(orgs)} organizations")

    if not orgs:
        raise typer.Abort()

    # Create a table of organizations
    columns = ["Name", "ID", "API"]
    if include_counts:
        columns.extend(["Networks", "Devices"])
    table = table_with_columns(columns, title="Organizations")

    with status_spinner("Gathering network and device counts"):
        for org in orgs:
            networks = devices = None
            if include_counts and org["api"]["enabled"]:
                try:
                    networks = dashboard.organizations.getOrganizationNetworks(
                        org["id"]
                    )
                    devices = dashboard.organizations.getOrganizationDevices(org["id"])
                except APIError:
                    console.print(f"Unable to access {org['name']}")

            row = [
                org["name"],
                org["id"],
                "[green]Enabled" if org["api"]["enabled"] else "[red]Disabled",
            ]
            if include_counts:
                row.extend(
                    [
                        str(len(networks)) if networks else "",
                        str(len(devices)) if devices else "",
                    ]
                )
            table.add_row(*row)

    console.print(table)


@app.command()
def network_health(organization_name: str):
    """
    Meraki network health (global alerts) for all network
    """
    org = find_org_by_name(organization_name)

    orgwide_health = []
    with status_spinner("Gathering health information"):
        networks = dashboard.organizations.getOrganizationNetworks(org["id"])
        for net in networks:
            health = dashboard.networks.getNetworkHealthAlerts(net["id"])
            for alert in health:
                alert["network_name"] = net["name"]
                orgwide_health.append(alert)

    console.print(
        table_network_health(
            health=orgwide_health,
            title=f"Health for {org['name']}",
            include_network_name=True,
        )
    )


@app.command()
def mx_uplinks(
    organization_name: str,
):
    """
    Return the uplink loss and latency for every MX in the organization
    """
    org = find_org_by_name(organization_name)
    with status_spinner("Gathering data"):
        networks = dashboard.organizations.getOrganizationNetworks(org["id"])
        devices = dashboard.organizations.getOrganizationDevices(org["id"])
        uplinks = dashboard.organizations.getOrganizationDevicesUplinksLossAndLatency(
            org["id"]
        )

    # Create network ID to name mapping
    network_map = {}
    for network in networks:
        network_map[network["id"]] = network["name"]

    # Create serial to device mapping
    device_map = {}
    for device in devices:
        device_map[device["serial"]] = device["name"]

    table = table_with_columns(
        ["Uplink", "IP", "Loss", "Latency", "Time"], first_column_name="Device"
    )

    for uplink in uplinks:
        table.add_row(
            f"{network_map[uplink['networkId']]} / {device_map[uplink['serial']]}",
            uplink["uplink"],
            uplink["ip"],
            f"{uplink['timeSeries'][-1]['lossPercent']}%"
            if uplink["timeSeries"][-1]["lossPercent"] is not None
            else "",
            f"{uplink['timeSeries'][-1]['latencyMs']}ms"
            if uplink["timeSeries"][-1]["latencyMs"] is not None
            else "",
            f"{uplink['timeSeries'][-1]['ts']}",
        )

    console.print(table)


@app.command()
def create(
    name: str,
    org_admin: Optional[List[str]] = typer.Option(
        None, help="Org admins in Name!Email format"
    ),
    claim_order: Optional[List[str]] = typer.Option(None),
):
    """
    Create a new Meraki organization
    """
    # Create organization
    try:
        org = dashboard.organizations.createOrganization(name=name)
        console.print(
            f"Created new organization [bold]{org['name']}[/bold]. ID: {org['id']}"
        )
    except APIError as err:
        console.print(f"Unable to create organization. {err.message}")
        raise typer.Abort()

    # Create org_admins if specified
    if org_admin:
        for admin in org_admin:
            name, email = admin.split("!")
            try:
                admin = dashboard.organizations.createOrganizationAdmin(
                    organizationId=org["id"], name=name, email=email, orgAccess="full"
                )
                console.print(f"Created new admin {email}.")
            except APIError as err:
                console.print(f"Unable to create admin {email}. {err.message}")
                raise typer.Abort()

    # Claim orders if specified
    if claim_order:
        try:
            dashboard.organizations.claimIntoOrganization(
                organizationId=org["id"], orders=claim_order
            )
            console.print(f"Claimed orders: {', '.join(claim_order)}")
        except APIError as err:
            console.print(f"Unable to claim order(s). {err.message}")


@app.command()
def saml(
    organization_name: str, enable: bool = typer.Option(None, "--enable/--disable")
):
    """
    Organization SAML status
    """
    # Get organization and print current status
    org = find_org_by_name(organization_name)
    saml_status = dashboard.organizations.getOrganizationSaml(organizationId=org["id"])[
        "enabled"
    ]
    console.print(
        f"SAML for [bold]{org['name']}[/bold] is currently"
        f" [bold]{'enabled' if saml_status else 'disabled'}."
    )

    if enable is not None:
        if saml_status == enable:
            # No change is requried
            console.print(
                " No change. SAML is already"
                f" [bold]{'enabled' if saml_status else 'disabled'}."
            )
        else:
            # Change API status
            with status_spinner("Updating organization"):
                saml_status = dashboard.organizations.updateOrganizationSaml(
                    organizationId=org["id"], enabled=enable
                )["enabled"]
            console.print(
                f" API is now [bold]{'enabled' if saml_status else 'disabled'}"
            )

    return saml_status


@app.command()
def create_saml_idp(organization_name: str, saml_x509_cert: str, saml_slo_logout: str):
    """
    Create a SAML IDP
    """
    # Find organization
    org = find_org_by_name(organization_name)

    # Create SAML IDP
    saml_idp = dashboard.organizations.createOrganizationSamlIdp(
        organizationId=org["id"],
        x509certSha1Fingerprint=saml_x509_cert,
        sloLogoutUrl=saml_slo_logout,
    )
    console.print(
        f"Created SAML IDP. Consumer URL is [bold]{saml_idp['consumerUrl']}[/bold]"
    )
    return saml_idp


@app.command()
def create_saml_role(
    organization_name: str,
    role: str,
    access: str = "full",
):
    """
    Create a SAML role
    """
    org = find_org_by_name(organization_name)
    new_role = dashboard.organizations.createOrganizationSamlRole(
        organizationId=org["id"], role=role, orgAccess=access
    )
    console.print(f"Created SAML role {new_role['role']}")


@app.command()
def claim_order(
    organization_name: str,
    order_number: Optional[List[str]] = typer.Option(...),
    claim_to_network_name: Optional[str] = typer.Option(None),
):
    """
    Claim an order into an organization
    """
    # Find organization or network
    if claim_to_network_name is None:
        org = find_org_by_name(org_name=organization_name)
        org_id = org["id"]
    else:
        net = find_network_by_name(
            org_name=organization_name, net_name=claim_to_network_name
        )
        org_id = net["organizationId"]

    # Attempt to claim order
    try:
        claim = dashboard.organizations.claimIntoOrganization(
            organizationId=org["id"], orders=order_number
        )
    except APIError as err:
        console.print(f"Unable to claim order(s). {err.message}")
        raise typer.Abort()

    console.print(f"Claimed orders: {', '.join(claim['orders'])}")

    # Find serials contained in this order number
    inventory = dashboard.organizations.getOrganizationInventoryDevices(
        organizationId=org_id
    )
    serials = []
    for device in inventory:
        if device["orderNumber"] in order_number:
            console.print(f" {device['model']} {device['serial']}")

            # Create a list of serials
            if claim_to_network_name is not None:
                if device["networkId"] is None:
                    serials.append(device["serial"])

    # Claim devices into networks if requested
    if claim_to_network_name is not None:
        dashboard.networks.claimNetworkDevices(networkId=net["id"], serials=serials)
        console.print(
            f"[green]{len(serials)} devices were added to [bold]{net['name']}"
        )


@app.command()
def api(
    organization_name: str, enable: bool = typer.Option(None, "--enable/--disable")
):
    """
    Organization API status
    """
    # Get organization and print current API status
    org = find_org_by_name(organization_name)
    api_status = org["api"]["enabled"]
    console.print(
        f"API for [bold]{org['name']}[/bold] is currently"
        f" [bold]{'enabled' if api_status else 'disabled'}."
    )

    # No change requested
    if enable is None:
        return api_status

    # No change required
    if api_status == enable:
        console.print(
            " No change. API is already"
            f" [bold]{'enabled' if api_status else 'disabled'}."
        )
        return api_status

    # Change API status
    with status_spinner("Updating organization"):
        org = dashboard.organizations.updateOrganization(
            organizationId=org["id"], name=org["name"], api={"enabled": enable}
        )
    console.print(f" API is now [bold]{'enabled' if enable else 'disabled'}")
    return enable


@app.command()
def create_ip_objects(
    organization_name: str, group_name: str = None, object: Optional[List[str]] = None
):
    """
    Create new IP objects within organization, optionally adding to specified group
    """
    if not object:
        console.print("No objects provided")
        raise typer.Abort()

    org = find_org_by_name(organization_name)

    # Iterate through each new object
    new_objects = []
    for obj in object:
        args = len(obj.split("!"))
        # Each port item should  be formatted as 'objectIP' or 'objectName!objectIP'
        if not 0 < args < 3:
            console.print("Incorect --object formatting")
            raise typer.Abort()

        if args == 2:
            # 'objectName!objectIP' format
            obj_name, obj_cidr = obj.split("!")
        elif args == 1:
            # 'objectIP' format
            obj_cidr = obj
            obj_name = obj.replace(".", "-")

        new_obj = api_req(
            f"organizations/{org['id']}/policyObjects",
            method="POST",
            json={
                "name": obj_name,
                "category": "network",
                "type": "cidr",
                "cidr": obj_cidr,
            },
        )
        new_objects.append(new_obj["id"])
        console.print(f"Created object named {obj_name} with IP {obj_cidr}")

    # Create a group with the new objects included
    if group_name:
        api_req(
            f"organizations/{org['id']}/policyObjects/groups",
            method="POST",
            json={
                "name": group_name,
                "category": "NetworkObjectGroup",
                "objectIds": new_objects,
            },
        )
        console.print(f"Created group named {group_name}")


@app.command()
def list_api_requests(organization_name: str):
    """
    List API requests for organization
    """
    org = find_org_by_name(organization_name)
    with status_spinner("Gathering API requests"):
        api_requests = dashboard.organizations.getOrganizationApiRequests(
            org["id"], total_pages="all", perPage=250
        )
    table = table_with_columns(
        ["Endpoint", "Response Code", "Source IP", "User Agent", "Time"],
        title=f"API Requests for {org['name']}",
    )
    for req in api_requests:
        table.add_row(
            f"{req['method']} {req['path']}{req['queryString']}",
            str(req["responseCode"]),
            req["sourceIp"],
            req["userAgent"][0:20],
            req["ts"],
        )

    console.print(table)
